from typing import Callable
import sys
from .parsers import parse_command_args
from .formatters import HelpFormatter, OutputFormatter
from .exceptions import CLIError

from ...log import Logger
from ...decorators import deprecated

logger = Logger({"log_line": {"format": []}})

class CLI:
    def __init__(self, name: str = "cli"):
        """The CLI class for creating interactive command-line applications.

        Args:
            name (str, optional): The name of the CLI. Defaults to "cli".
        """
        self.name = name
        self.commands: dict[str, Callable] = {}

    def command(self, func: Callable) -> Callable:
        """Decorator to register a function as a CLI command."""
        self.commands[func.__name__] = func
        return func

    def run(self, args: list[str] = None) -> None:
        """Run the CLI application."""
        if not args:
            args = sys.argv[1:]

        if not args or args[0] in ("-h", "--help"):
            self.print_help()
            return

        command_name = args[0]
        if command_name in self.commands:
            try:
                # fetch command
                command = self.commands[command_name]
                # parse arguments
                parsed_args = parse_command_args(command, args[1:])
                # change terminal title
                print(f"\x1b]0;{self.name} {command_name}\x7", end="")
                # run command
                result = command(**parsed_args)
                if result is not None:
                    logger.info(result)
            except CLIError as e:
                print(f"Error: {str(e)}")
                self.print_command_help(command_name)
        else:
            print(f"Unknown command: {command_name}")
            self.print_help()

    def print_help(self) -> None:
        """Print help information for the entire CLI."""
        help_text = HelpFormatter.format_cli_help(self.name, self.commands)
        print(help_text)

    def print_command_help(self, command_name: str) -> None:
        """Print help information for a specific command."""
        if command_name in self.commands:
            help_text = HelpFormatter.format_command_help(
                command_name, self.commands[command_name]
            )
            print(help_text)
        else:
            print(f"Unknown command: {command_name}")

    @deprecated(expected_removal="v1.0.0")
    def echo(self, message: any) -> None:
        """Print a formatted message to the console. Works with lists, tuples, and dictionaries. Other formats are printed as is."""
        formatted_output = OutputFormatter.format_output(message)
        print(formatted_output)



from zenif.cli import CLI, req, opt, Prompt as p, install_setup
from zenif.schema import (
    Schema,
    BooleanF,
    StringF,
    IntegerF,
    ListF,
    Length,
    Value,
    Email,
    NotEmpty,
)
import os

cli = CLI(name="cli")

install_setup(cli, os.path.abspath(__file__))


@cli.command
@req("name", help="Name to greet")
@opt("--greeting", default="Hello", help="Greeting to use")
@opt("--shout", is_flag=True, help="Print in uppercase")
def greet(name: str, greeting: str, shout: bool = False):
    """Greet a person."""
    message = f"{greeting}, {name}!"
    if shout:
        message = message.upper()
    return message


@cli.command
def test_prompts():
    """Test all available prompts"""

    class OddOrEven:
        def __init__(self, parity: str = "even"):
            self.parity = 1 if parity == "odd" else 0

        def __call__(self, value):
            if value % 2 != self.parity:
                raise ValueError(
                    f"Must be an {'even' if self.parity ==
                                  0 else 'odd'} number."
                )

    # clear the screen
    os.system("cls" if os.name == "nt" else "clear")

    schema = Schema(
        are_you_sure=BooleanF().name("continue"),
        name=StringF().name("name").has(Length(min=3, max=50)),
        password=StringF().name("password").has(NotEmpty()),
        age=IntegerF()
        .name("age")
        .has(Value(min=18, max=120))
        .has(OddOrEven(parity="odd")),
        interests=ListF()
        .name("interests")
        .item_type(StringF())
        .has(Length(min=3, err="Select a minimum of 3 interests.")),
        fav_interest=StringF().name("fav_interest"),
        email=StringF().name("email").has(Email()),
    ).all_optional()

    for i in range(4):
        print(i + 1)

    if (
        not p.confirm("Are you sure you want to continue?", schema, "are_you_sure")
        .default(True)
        .ask()
    ):
        return
    # name = p.text("Enter your name", schema, "name").ask()
    # email = p.text("Enter your email", schema, "email").ask()
    # password = p.password("Enter your password", schema, "password").peeper().ask()
    # date = p.date("Enter your date of birth").month_first().show_words().ask()
    # age = p.number("Enter your age", schema, "age").ask()
    editor = p.editor("Enter your hacker code").language("py").ask()
    interests = p.checkbox(
        "Select your interests",
        ["Reading", "Gaming", "Sports", "Cooking", "Travel"],
        schema,
        "interests",
    ).ask()
    fav_interest = p.choice(
        "Select your favorite interest",
        interests,
        schema,
        "fav_interest",
    ).ask()

    # print(f"{name=}")
    # print(f"{email=}")
    # print(f"{password=}")
    # print(f"{date=}")
    # print(f"{age=}")
    print(f"{editor=}")
    print(f"{interests=}")
    print(f"{fav_interest=}")


if __name__ == "__main__":
    cli.run()



from __future__ import annotations

from .core import SchemaField

from enum import Enum
from datetime import datetime
from ast import literal_eval


class StringF(SchemaField[str]):
    def coerce(self, value: any) -> str:
        return str(value)


class IntegerF(SchemaField[int]):
    def coerce(self, value: any) -> int:
        return int(float(value))


class FloatF(SchemaField[float]):
    def coerce(self, value: any) -> float:
        return float(value)


class BooleanF(SchemaField[bool]):
    def coerce(self, value: any) -> bool:
        if isinstance(value, str):
            return value.lower() in ("true", "1", "yes", "on")
        return bool(value)


class DateF(SchemaField[datetime]):
    def coerce(self, value: any) -> datetime:
        if isinstance(value, str):
            return datetime.fromisoformat(value)
        return value


class EnumF(SchemaField[Enum]):
    def __init__(self):
        super().__init__()
        self._enum_class: type[Enum] | None = None

    def enum_class(self, enum_class: type[Enum]) -> EnumF:
        self._enum_class = enum_class
        return self

    def coerce(self, value: any) -> Enum:
        if self._enum_class is None:
            raise ValueError("Enum class not set")
        if isinstance(value, str):
            return self._enum_class[value.upper()]
        return self._enum_class(value)


class ListF(SchemaField[list]):
    def __init__(self):
        super().__init__()
        self._item_type: SchemaField | None = None

    def item_type(self, item_type: SchemaField) -> ListF:
        self._item_type = item_type
        return self

    def coerce(self, value: any) -> list:
        if isinstance(value, str):
            value = literal_eval(value)
        if not isinstance(value, list):
            value = [value]
        if self._item_type:
            return [self._item_type.coerce(item) for item in value]
        return value


class DictF(SchemaField[dict]):
    def __init__(self):
        super().__init__()
        self._key_type: SchemaField | None = None
        self._value_type: SchemaField | None = None

    def key_type(self, key_type: SchemaField) -> DictF:
        self._key_type = key_type
        return self

    def value_type(self, value_type: SchemaField) -> DictF:
        self._value_type = value_type
        return self

    def coerce(self, value: any) -> dict:
        if isinstance(value, str):
            value = literal_eval(value)
        if not isinstance(value, dict):
            raise ValueError("Cannot coerce to dict")
        if self._key_type and self._value_type:
            return {
                self._key_type.coerce(k): self._value_type.coerce(v)
                for k, v in value.items()
            }
        return value



from .core import Schema
from .fields import (
    StringF,
    FloatF,
    DateF,
    DictF,
    EnumF,
    ListF,
    BooleanF,
    IntegerF,
    SchemaField,
)
from .validators import Validator, Length, Value, Regex, Email, Date, URL, NotEmpty, Alphanumeric

__all__ = [
    "Schema",
    # Fields
    "SchemaField",  # base class
    "StringF",
    "IntegerF",
    "FloatF",
    "BooleanF",
    "ListF",
    "DictF",
    "EnumF",
    "DateF",
    # Validators
    "Validator",  # base class
    "Length",
    "Value",
    "Regex",
    "Email",
    "Date",
    "URL",
    "NotEmpty",
    "Alphanumeric",
]



from typing import Callable
import argparse
from .exceptions import CLIError


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise CLIError(message)


class CommandParser:
    def __init__(self, command: Callable):
        self.command = command
        self.parser = ArgumentParser(description=command.__doc__)
        self._add_arguments()

    def _add_arguments(self):
        if hasattr(self.command, "_arguments"):
            for args, kwargs in self.command._arguments:
                self._add_argument(*args, **kwargs)

    def _add_argument(self, *args, **kwargs):
        # Remove custom parameters that argparse doesn't understand
        is_flag = kwargs.pop("is_flag", False)
        is_option = kwargs.pop("is_option", False)

        if is_flag:
            kwargs["action"] = "store_true"
            kwargs.setdefault("default", False)

        if is_option and not args[0].startswith("-"):
            args = (f"--{args[0]}",) + args[1:]

        self.parser.add_argument(*args, **kwargs)

    def parse_args(self, args: list[str]) -> dict[str, any]:
        try:
            parsed_args = self.parser.parse_args(args)
            return vars(parsed_args)
        except CLIError as e:
            print(f"Error: {str(e)}")
            self.parser.print_help()
            return {}


def parse_command_args(command: Callable, args: list[str]) -> dict[str, any]:
    parser = CommandParser(command)
    return parser.parse_args(args)



from zenif.log import Logger
import sys

# Initialize a Logger instance
l = Logger()

fstream = l.stream.file
nstream = l.stream.normal

group = l.fhgroup(
    fstream.add("log1.log"),
    l.fhgroup(fstream.add("log2.log"), fstream.add("log3.log")),
    l.fhgroup(fstream.add("log4.log")),
)

group.reset()
group.modify(
    ruleset={
        "timestamps": {
            "always_show": True,
            "use_utc": True,
        },
        "log_line": {"format": "noalign"},
    }
)


def superlongfunctionname():
    l.error("Could", "not", "run function")


# Tests truncation of long function names
superlongfunctionname()

l.info(l.ruleset.dict.all)  # -> l.__rules
l.info(l.ruleset.dict.stacking)  # -> l.__rules["stacking"]
l.info(l.ruleset.stacking)  # -> StackingRules object
l.info(l.ruleset.stacking.enabled)  # -> l.__rules["stacking"]["enabled"]

# Sample logging tests
l.info("a\n", 3, 4, 4, "hello world 3")
l.info("a", 3, 4, 4, "hello world 3", sep="")
l.info(
    "a\n\n\n\n", 3, 4, 4, "hello world 3", ["a\n\n\n\n", 3, 4.0, 4.2, "hello world 3"]
)
l.info("a", 3, 4, 4, "hello world 3", sep=" ")
l.info("a", 3, 4, 4, "hello world 3", sep="_")
l.info("a")
l.info("ajkfhgfhkgdjhd")
l.info(345)

l.debug(3.0)
l.debug(3.0, {})
l.debug("Dictionary:", {}, sep=" ")

nstream.modify(sys.stdout, ruleset={"timestamps": {"always_show": True}})


x = 33
l.lethal(
    {
        "abcdef": "abcdef",
        "l": ["l", ["l", ["l", ["l", ["l", ["l", []]]]]]],
        "mnopqr": "mnopqr",
        "stuvwx": "stuvwx",
        "hi": (x, l),
    }
)

l.error("hi", "bye")
l.warning("The given tuple is not a valid RGB:", (132, 204, 256))

# Remove the log.log file stream
fstream.remove("log1.log")

l.warning("The given tuple is not a valid RGB:", (132, 204, 256))
l.warning("The given tuple is not a valid RGB:", (132, 204, 256))
l.warning("The given tuple is not a valid RGB:", (132, 204, 256))


# Create a file handler group
fh_group = l.fhgroup(
    fstream.add("log1.log"), fstream.add("log2.log"), fstream.add("log3.log")
)

l.debug("Added groups:", fh_group.file_paths)


# Modify all files in the group


fh_group.modify({"timestamps": {"use_utc": False}})

# Add another file to the group

l.debug("Added groups:", fh_group.add("log4.log"))


# # Reset all files in the group
# fh_group.reset()

# Remove a specific file from the group
fh_group.remove("log2.log", l.fhgroup("log3.log", l.fhgroup("log1.log")))

l.debug("Logs 1-3 won't get this message!")

# # Remove all files in the group
# fh_group.remove_all()

import sys
from io import StringIO

sh_group1 = l.shgroup(sys.stdout)
sh_group2 = l.shgroup(sys.stderr)
custom_stream = StringIO()
combined_sh_group = l.shgroup(sh_group1, sh_group2, custom_stream)

l.error("hello")
l.error("hello")
l.error("hello")



from __future__ import annotations
from typing import Callable, Generic, TypeVar

T = TypeVar("T")


class Condition:
    def __init__(self, condition: Callable[[dict], bool], error_message: str):
        self.condition = condition
        self.error_message = error_message

    def check(self, data: dict) -> bool:
        return self.condition(data)


class Validator:
    def __init__(self, err: str | None = None):
        if err:
            self.err = f"{err}{"" if err.endswith(".") else "."}"
        else:
            self.err = ""

    def __call__(self, value: any) -> any:
        self.validate(value=value)

    def validate(self, value: any):
        try:
            self._validate(value)
        except ValueError as e:
            if self.err:
                raise ValueError(self.err)
            else:
                raise e
            raise e

    def _validate(self, value: any):
        raise NotImplementedError()


class SchemaField(Generic[T]):
    def __init__(self):
        self._name: str | None = None
        self._default: any | None = None

        self.validators: list[Validator] = []

        self.is_required: bool = True

        self.condition: Condition | None = None
        self.pre_transform: Callable[[any], any] | None = None
        self.post_transform: Callable[[T], any] | None = None

    def name(self, name: str) -> SchemaField[T]:
        self._name = name
        return self

    def has(self, validator: Validator) -> SchemaField[T]:
        self.validators.append(validator)
        return self

    def when(
        self, condition: Callable[[dict], bool], error_message: str
    ) -> "SchemaField[T]":
        self.condition = Condition(condition, error_message)
        return self

    def default(self, value: T | Callable[[], T]) -> SchemaField[T]:
        self._default = value
        self.is_required = False
        return self

    def optional(self) -> SchemaField[T]:
        self.is_required = False
        return self

    def pre(self, func: Callable[[any], any]) -> "SchemaField[T]":
        self.pre_transform = func
        return self

    def post(self, func: Callable[[T], any]) -> "SchemaField[T]":
        self.post_transform = func
        return self

    def coerce(self, value: any) -> T:
        return value  # Default implementation, subclasses should override if needed


class Schema:
    def __init__(self, **fields: SchemaField):
        """A class for validating and coercing data based on a schema."""

        self.fields = fields
        self._strict = False
        self._all_optional = False

    def strict(self, value: bool = True) -> Schema:
        """Set strict mode to True or False."""
        self._strict = value
        return self

    def all_optional(self) -> Schema:
        """Mark all fields as optional."""
        self._all_optional = True
        return self

    def validate(self, data: dict) -> tuple[bool, dict[str, list[str]], dict]:
        """Validate data against the schema.

        Args:
            data (dict): The data to validate.

        Raises:
            SyntaxError: If a field name does not end with "F".

        Returns:
            tuple[bool, dict[str, list[str]], dict]: A tuple containing a boolean indicating whether the data is valid, a dictionary of field errors, and finally a dictionary of coerced data.
        """
        is_valid = True
        errors: dict[str, list[str]] = {}
        coerced_data = {}

        for field_name, field in self.fields.items():
            if not field.__class__.__name__.endswith("F"):
                raise SyntaxError(
                    f'Field {field.__class__.__name__} name must end with "F".'
                )
            if field.condition:
                if not field.condition.check(data):
                    continue  # Skip this field if the condition is not met
            if field_name not in data:
                if field.is_required and not self._all_optional:
                    is_valid = False
                    errors[field_name] = ["This field is required."]
                elif field._default is not None:
                    coerced_data[field_name] = (
                        field._default() if callable(field._default) else field._default
                    )
            else:
                try:
                    value = data[field_name]

                    if not self._strict:
                        value = field.coerce(value)

                    if field.pre_transform:
                        value = field.pre_transform(value)

                    field_errors = []
                    for validator in field.validators:
                        try:
                            validator(value)
                        except ValueError as e:
                            is_valid = False
                            field_errors.append(str(e))

                    if field.post_transform:
                        value = field.post_transform(value)

                    if field_errors:
                        errors[field_name] = field_errors
                    else:
                        coerced_data[field_name] = value
                except Exception as e:
                    is_valid = False
                    errors[field_name] = [str(e)]

        if self._strict:
            extra_fields = set(data.keys()) - set(self.fields.keys())
            if extra_fields:
                is_valid = False
                errors["__extra__"] = [f"Unexpected fields: {', '.join(extra_fields)}"]

        return is_valid, errors, coerced_data



from zenif.cli import CLI, req, opt, Prompt, install_setup
from zenif.log import Logger
from zenif.decorators import retry, cache
import os
import random

cli = CLI(name="zenif-demo")
logger = Logger(ruleset={"log_line": {"format": "simple"}})

# Add the install command to your CLI
install_setup(cli, os.path.abspath(__file__))


@cli.command
@req("name", help="Your name")
@opt("--greeting", default="Hello", help="Greeting to use")
def greet(name: str, greeting: str):
    """Greet the user"""
    logger.info(f"{greeting}, {name}!")


@cli.command
@retry(max_retries=8, delay=1.0)
def flaky_operation():
    """Demonstrate a flaky operation with retry on exception"""
    logger.info("Attempting a flaky operation...")
    if random.random() < 0.7:  # 70% chance of failure
        logger.error("Randon failure occurred! Retrying...")
        raise Exception("Random failure occurred!")
    logger.success("Flaky operation succeeded!")


@cli.command
@req("n", help="Calculate the nth Fibonacci number")
@cache
def fibonacci(n: int):
    """Calculate the nth Fibonacci number"""
    n = int(n)
    if n > 488:
        logger.error("n must less than or equal to 488.")
        return
    if n <= 1:
        return n
    else:
        return fibonacci(n - 1) + fibonacci(n - 2)


@cli.command
def interactive_prompt():
    """Demonstrate interactive prompts"""
    name = Prompt.text("What's your name?").ask()
    age = Prompt.number("How old are you?").min(0).max(120).ask()
    likes_python = Prompt.confirm("Do you like Python?").ask()
    favorite_color = Prompt.choice(
        "What's your favorite color?", choices=["Red", "Green", "Blue", "Yellow"]
    ).ask()

    logger.info(f"Name: {name}")
    logger.info(f"Age: {age}")
    logger.info(f"Likes Python: {likes_python}")
    logger.info(f"Favorite color: {favorite_color}")


@cli.command
@opt("--verbose", is_flag=True, help="Enable verbose logging")
def log_demo(verbose: bool):
    """Demonstrate different log levels"""
    if verbose:
        logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.success("This is a success message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")


if __name__ == "__main__":
    cli.run()



from ...schema import Schema
from ...constants import Keys, Cursor

import signal
import sys
from colorama import init, Fore, Style

init(autoreset=True)


class BasePrompt:
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        self.message = message
        self.schema = schema
        self.id = id
        if schema and not id:
            raise ValueError("You must have an ID in order to use a schema.")
        if schema and id:
            self.field = schema.fields.get(id)
            if not self.field:
                raise ValueError(f"Field '{id}' not found in the schema.")
        else:
            self.field = None

    def validate(self, value):
        try:
            if self.schema and self.id:
                is_valid, errors, _ = self.schema.validate({self.id: value})
                if not is_valid:
                    return errors.get(self.id, ["Invalid input"])[0].rstrip(".")
            elif self.field:
                self.field.validate(value)
            return None
        except ValueError as e:
            return str(e)

    @staticmethod
    def _get_key():
        def handle_interrupt(signum, frame):
            raise KeyboardInterrupt()

        if sys.platform.startswith("win"):
            import msvcrt

            # Set up the interrupt handler
            signal.signal(signal.SIGINT, handle_interrupt)

            try:
                while True:
                    if msvcrt.kbhit():
                        char = msvcrt.getch().decode("utf-8")
                        if char == Keys.CTRLC:  # Ctrl+C
                            raise KeyboardInterrupt()
                        return char
            finally:
                # Reset the interrupt handler
                signal.signal(signal.SIGINT, signal.SIG_DFL)

        else:
            import termios
            import tty

            fd = sys.stdin.fileno()
            old_settings = termios.tcgetattr(fd)
            try:
                tty.setraw(fd)
                # Set up the interrupt handler
                signal.signal(signal.SIGINT, handle_interrupt)

                while True:
                    char = sys.stdin.read(1)
                    if char == Keys.CTRLC:  # Ctrl+C
                        raise KeyboardInterrupt()
                    if char == Keys.ESCAPE:
                        # Handle escape sequences (e.g., arrow keys)
                        next_char = sys.stdin.read(1)
                        if next_char == "[":
                            last_char = sys.stdin.read(1)
                            return f"\x1b[{last_char}"
                    return char
            finally:
                # Reset terminal settings and interrupt handler
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                signal.signal(signal.SIGINT, signal.SIG_DFL)

    @staticmethod
    def _print_prompt(
        prompt: str = "",
        value: str = "",
        default: str | None = None,
        options: list[str] | None = None,
        default_option: str | None = None,
        error: str | None = None,
    ):
        sys.stdout.write(
            f"{Cursor.lclear()}\r{Fore.GREEN}? {Fore.CYAN}{prompt}{Fore.RESET}"
        )
        if default and not options:
            sys.stdout.write(f" {Fore.CYAN}{Style.DIM}({default}){Style.RESET_ALL}")
        if options:
            if default_option:
                options[
                    [option.lower() for option in options].index(default_option.lower())
                ] = options[
                    [option.lower() for option in options].index(default_option.lower())
                ].upper()
            if len(options) == 2:
                sys.stdout.write(
                    f" {Fore.CYAN}{Style.DIM}[{options[0]}/{options[1]}]{Style.RESET_ALL}"
                )
            else:
                sys.stdout.write(
                    f" {Fore.CYAN}{Style.DIM}[{"".join(options)}]{Style.RESET_ALL}"
                )
        sys.stdout.write(f"{Fore.CYAN} {Fore.YELLOW}{value}")
        if error:
            sys.stdout.write(f"  {Fore.RED}{error}{Cursor.left(2 + len(error))}")

        sys.stdout.flush()



from __future__ import annotations

from .core import Validator

from re import match

inf = float("inf")


class Length(Validator):
    """Compares whether the length of the value is within the given range."""

    def __init__(
        self, min: int | None = None, max: int | None = None, err: str | None = None
    ):
        super().__init__(err)
        self.min = min if min is not None else -inf
        self.max = max if max is not None else inf

    def _validate(self, value: any):
        if value is None:
            raise ValueError(f"Value is of None type.")
        if len(value) < self.min:
            raise ValueError(f"Minimum length is {self.min}.")
        if len(value) > self.max:
            raise ValueError(f"Maximum length is {self.max}.")


class Value(Validator):
    """Compares whether the value is within the given range."""

    def __init__(
        self, min: int | None = None, max: int | None = None, err: str | None = None
    ):
        super().__init__(err)
        self.min = min if min is not None else -inf
        self.max = max if max is not None else inf

    def _validate(self, value: any):
        if value is None:
            raise ValueError(f"Value is of None type.")
        if value < self.min:
            raise ValueError(f"Minimum value is {self.min}.")
        if value > self.max:
            raise ValueError(f"Maximum value is {self.max}.")


class Regex(Validator):
    """Matches the given pattern to the value."""

    def __init__(self, pattern: str, err: str | None = None):
        super().__init__(err)
        self.pattern = pattern

    def _validate(self, value: any):
        if not match(self.pattern, str(value)):
            raise ValueError(f"Value does not match pattern.")


class Email(Regex):
    """Extends the Regex() validator with the official RFC 5322 email regular expression."""

    def __init__(self, err: str | None = None):
        if err is None:
            err = "Invalid email address."
        super().__init__(
            r"""(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])""",
            err,
        )


class Alphanumeric(Regex):
    """Ensures the value is alphanumeric."""

    def __init__(self, err: str | None = None):
        if err is None:
            err = "Value must be alphanumeric."
        super().__init__(r"^[a-zA-Z0-9]+$", err)


class URL(Regex):
    """Validates that the value is a valid URL."""

    def __init__(self, err: str | None = None):
        if err is None:
            err = "Invalid URL."
        super().__init__(r"^(https?|ftp)://[^\s/$.?#].[^\s]*$", err)


class Date(Regex):
    """Validates that the value matches the YYYY-MM-DD format."""

    def __init__(self, err: str | None = None):
        if err is None:
            err = "Invalid date format. Expected format is YYYY-MM-DD."
        super().__init__(r"^\d{4}-\d{2}-\d{2}$", err)


class NotEmpty(Validator):
    """Validates that the value is not empty."""

    def __init__(self, err: str | None = None):
        super().__init__(err)

    def _validate(self, value: any):
        if not value:
            raise ValueError("Value cannot be empty.")



from os import stat, chmod, path, unlink
from stat import S_IEXEC
from typing import Callable
from .core import CLI
from .decorators import opt
from subprocess import run, CalledProcessError
from tempfile import NamedTemporaryFile


def install_setup(cli: CLI, script_path: str) -> Callable:
    """
    Install a setup command for a Zenif CLI applet.

    This function creates a new command that generates a shell script to install
    the CLI applet as a Zsh command.

    Args:
        cli (CLI): The Zenif CLI instance to add the command to.
        script_path (str): The file to run.

    Returns:
        Callable: The setup function that was added to the CLI.
    """

    @cli.command
    @opt("--alias", default=cli.name, help="Alias for the command")
    def setup(alias: str):
        """Install this script as a Zsh command"""
        install_script_content = f"""#!/bin/zsh
if [[ -f ~/.zshrc ]]; then
    if ! grep -q "{alias}()" ~/.zshrc; then
        echo '\n# Auto-generated by Zenif CLI' >> ~/.zshrc
        echo '{alias}() {{' >> ~/.zshrc
        echo '    python {script_path} "$@"' >> ~/.zshrc
        echo '}}' >> ~/.zshrc
        echo "Command \x1b[44m\x1b[30m {alias} \x1b[0m has been added to your .zshrc file.\n"
        echo "Run the following to finish the installation:"
        echo "\nsource ~/.zshrc   \x1b[90mupdate your configuration\x1b[0m"
        echo   "{alias} -h{(13-len(alias))*" "}  \x1b[90mdisplay help information\x1b[0m"
    else
        echo "The command \x1b[44m\x1b[30m {alias} \x1b[0m is already in your .zshrc file."
    fi
else
    echo "Error: ~/.zshrc file not found."
fi

# Delete this script
rm -f "$0"
"""

        with NamedTemporaryFile(mode="w", suffix=".sh", delete=False) as temp_script:
            temp_script.write(install_script_content)
            temp_script_path = temp_script.name

        # Make the script executable
        st = stat(temp_script_path)
        chmod(temp_script_path, st.st_mode | S_IEXEC)

        print(f"Executing installation script for '{alias}' command...")

        try:
            # Execute the script
            result = run(
                ["zsh", temp_script_path], check=True, text=True, capture_output=True
            )
            print(f"\n{result.stdout}")
            if result.stderr:
                print(f"\n{result.stderr}")
        except CalledProcessError as e:
            print(f"Installation failed with error: {e}")
            print(e.stderr)
        finally:
            try:
                if "temp_script_path" in locals() and path.exists(temp_script_path):
                    unlink(temp_script_path)
                    print("Installation script has been cleaned up.")
                else:
                    print("No temporary script to clean up; already cleaned up.")
            except Exception as e:
                print(f"Failed to delete temporary script: {e}")

    return setup



from typing import Callable


def req(*args, **kwargs):
    """Decorator to add a required argument to a CLI command."""

    # for each arg and kwarg, set the help to lowercase
    kwargs["help"] = kwargs.get("help", "").lower()

    def decorator(func: Callable) -> Callable:
        if not hasattr(func, "_arguments"):
            func._arguments = []
        func._arguments.append((args, kwargs))
        return func

    return decorator


def opt(*args, **kwargs):
    """Decorator to add an optional argument or flag to a CLI command."""
    kwargs["is_option"] = True
    return req(*args, **kwargs)



from .base import BasePrompt
from ...schema import Schema
from ...constants import Keys, Cursor

from colorama import init, Fore, Style

init(autoreset=True)


class CheckboxPrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        choices: list[str],
        schema: Schema | None = None,
        id: str | None = None,
    ):

        super().__init__(message, schema, id)
        self.choices = choices

    def ask(self) -> list[str]:
        """Prompt the user for input."""
        selected = [False] * len(self.choices)
        current = 0

        controls = "↑/↓ to navigate, Space to select, Enter to confirm"

        print()

        i = True
        while True:
            for i, (choice, is_selected) in enumerate(zip(self.choices, selected)):
                if i == current:
                    print(f"{Fore.YELLOW}{Style.DIM}X{Style.NORMAL}", end="")
                else:
                    print(f"{Fore.YELLOW} ", end="")
                print(
                    f"\r{f"{Fore.YELLOW}{"\x1b[4m" if i == current else ""}X\x1b[0m" if is_selected else Cursor.right(1)} {Fore.YELLOW}{Style.DIM}{choice}{Fore.RESET}"
                )

            if i:
                i = False
                result = [
                    choice
                    for choice, is_selected in zip(self.choices, selected)
                    if is_selected
                ]
                error = self.validate(result)

                print(Cursor.up(len(self.choices) + 2), end="")
                self._print_prompt(self.message, error=f"{error if error else ""}\n")
                print(
                    f"\r{Fore.RESET}{Style.DIM}  {controls}{Cursor.down(len(self.choices))}"
                )

            key = self._get_key()
            if key == " ":  # Space
                selected[current] = not selected[current]

            result = [
                choice
                for choice, is_selected in zip(self.choices, selected)
                if is_selected
            ]
            error = self.validate(result)

            print(Cursor.up(len(self.choices) + 2), end="")
            self._print_prompt(self.message, error=f"{error if error else ""}\n")
            print(
                f"\r{Fore.RESET}{Style.DIM}  {controls}{Cursor.down(len(self.choices))}"
            )

            if key == Keys.ENTER and not error:
                for _ in range(len(self.choices) + 2):
                    print(Cursor.up(1) + Cursor.lclear(), end="")
                self._print_prompt(
                    self.message,
                    (
                        ", ".join(map(str, result[:-1])) + f", and {result[-1]}"
                        if len(result) > 1
                        else str(result[0])
                    ),
                )
                print()  # Move to next line
                return result
            elif key == Keys.UP and current > 0:  # Up arrow
                current -= 1
            elif key == Keys.DOWN and current < len(self.choices) - 1:  # Down arrow
                current += 1

            print(Cursor.up(len(self.choices) + 1))  # Move cursor up to redraw choices



from .applets import CLI, req, opt, install_setup

from .prompt import Prompt


__all__ = [
    # CLI()
    "CLI",
    # CLI().command decorators
    "req",
    "opt",
    # setup command installer
    "install_setup",
    # interactive prompt utils
    "Prompt",
]



from textwrap import dedent


class HelpFormatter:
    @staticmethod
    def format_command_help(command_name: str, command: any) -> str:
        """Format help text for a single command."""
        help_text = [f"Command: {command_name}"]
        if command.__doc__:
            help_text.append(dedent(command.__doc__).strip())

        if hasattr(command, "_arguments"):
            help_text.append("\nArguments:")
            for args, kwargs in command._arguments:
                arg_help = f"  {', '.join(args)}"
                if "help" in kwargs:
                    arg_help += f": {kwargs['help']}"
                help_text.append(arg_help)

        return "\n".join(help_text)

    @staticmethod
    def format_cli_help(cli_name: str, commands: dict[str, any]) -> str:
        """Format help text for the entire CLI application."""
        help_text = [f"usage: {cli_name} <command> [args]", "\navailable commands:"]
        for name, command in commands.items():
            doc = command.__doc__.strip() if command.__doc__ else "no description"
            help_text.append(f"  {name}: {doc.split('\n')[0]}")
        return "\n".join(help_text)


class OutputFormatter:
    @staticmethod
    def format_output(output: any) -> str:
        """Format output for display."""
        if isinstance(output, (list, tuple)):
            return "\n".join(map(str, output))
        elif isinstance(output, dict):
            return "\n".join(f"{k}: {v}" for k, v in output.items())
        else:
            return str(output)
